from .main import ChatGPT
from importlib import metadata

__all__ = ["ChatGPT"]

try:
    __version__ = metadata.version("chatgpt_anon")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0"
__author__ = "Smartwa"
__repo__ = "https://github.com/Simatwa/chatgpt-anon"
__info__ = "Intreract with ChatGPT anonymously."



# Works at the terminal

import click
import cmd
import rich
import os
import sys
import re
import getpass
from rich.panel import Panel
from rich.style import Style
from rich.markdown import Markdown
from rich.live import Live
from rich.prompt import Prompt
from rich.console import Console
from rich.table import Table
from .main import ChatGPT
from time import sleep
import logging
import dotenv
import datetime
import json
import clipman
from functools import wraps
from threading import Thread as thr
from . import __repo__, __version__, __author__, __info__
from .utils import error_handler, get_message
from typing import Iterator

getExc = lambda e: e.args[1] if len(e.args) > 1 else str(e)

rich_code_themes = ["monokai", "paraiso-dark", "igor", "vs", "fruity", "xcode"]

logging.basicConfig(
    format="%(asctime)s - %(levelname)s : %(message)s ",  # [%(module)s,%(lineno)s]", # for debug purposes
    datefmt="%H:%M:%S",
    level=logging.INFO,
)

try:
    clipman.init()
except Exception as e:
    logging.debug(f"Dropping clipman in favor of pyperclip - {getExc(e)}")
    import pyperclip

    clipman.set = pyperclip.copy
    clipman.get = pyperclip.paste


def stream_output(
    iterable: Iterator,
    title: str = "",
    is_markdown: bool = True,
    style: object = Style(),
    transient: bool = False,
    title_generator: object = None,
    title_generator_params: dict = {},
    code_theme: str = "monokai",
    vertical_overflow: str = "ellipsis",
) -> None:
    """Stdout streaming response

    Args:
        iterable (Iterator): Iterator containing contents to be stdout
        title (str, optional): Content title. Defaults to ''.
        is_markdown (bool, optional): Flag for markdown content. Defaults to True.
        style (object, optional): `rich.style` instance. Defaults to Style().
        transient (bool, optional): Flag for transient. Defaults to False.
        title_generator (object, optional): Function for generating title. Defaults to None.
        title_generator_params (dict, optional): Kwargs for `title_generator` function. Defaults to {}.
        code_theme (str, optional): Theme for styling codes. Defaults to `monokai`
        vertical_overflow (str, optional): Response rendering vertical overflow behavior. Defaults to ellipsis.
    """
    render_this = ""
    with Live(
        render_this,
        transient=transient,
        refresh_per_second=16,
        vertical_overflow=vertical_overflow,
    ) as live:
        for entry in iterable:
            render_this += entry
            live.update(
                Panel(
                    Markdown(entry, code_theme=code_theme) if is_markdown else entry,
                    title=title,
                    style=style,
                )
            )
        if title_generator:
            title = title_generator(**title_generator_params)
            live.update(
                Panel(
                    Markdown(entry, code_theme=code_theme) if is_markdown else entry,
                    title=title,
                    style=style,
                )
            )


def stream_console_output(
    iterable: Iterator,
    title: str = "",
    is_markdown: bool = True,
    style: object = Style(),
    transient: bool = False,
    title_generator: object = None,
    title_generator_params: dict = {},
    code_theme: str = "monokai",
    vertical_overflow: str = "ellipsis",
) -> None:
    """Stdout streaming response without frame

    Args:
        iterable (Iterator): Iterator containing contents to be stdout
        title (str, optional): Content title. Defaults to ''.
        is_markdown (bool, optional): Flag for markdown content. Defaults to True.
        style (object, optional): `rich.style` instance. Defaults to Style().
        transient (bool, optional): Flag for transient. Defaults to False.
        title_generator (object, optional): Function for generating title. Defaults to None.
        title_generator_params (dict, optional): Kwargs for `title_generator` function. Defaults to {}.
        code_theme (str, optional): Theme for styling codes. Defaults to `monokai`
        vertical_overflow (str, optional): Response rendering vertical overflow behavior. Defaults to ellipsis.
    """
    console = Console(style=style)
    with Live(
        console=console,
        transient=transient,
        refresh_per_second=16,
        vertical_overflow=vertical_overflow,
    ) as live:
        for entry in iterable:
            live.update(
                Markdown(entry, code_theme=code_theme) if is_markdown else entry,
            )


class busy_bar:
    querying = None
    __spinner = (
        (),
        ("-", "\\", "|", "/"),
        (
            "█■■■■",
            "■█■■■",
            "■■█■■",
            "■■■█■",
            "■■■■█",
        ),
        ("⣾ ", "⣽ ", "⣻ ", "⢿ ", "⡿ ", "⣟ ", "⣯ ", "⣷ "),
    )
    spin_index = 0
    sleep_time = 0.1

    @classmethod
    def __action(
        cls,
    ):
        while cls.querying:
            for spin in cls.__spinner[cls.spin_index]:
                print(" " + spin, end="\r", flush=True)
                if not cls.querying:
                    break
                sleep(cls.sleep_time)

    @classmethod
    def start_spinning(
        cls,
    ):
        try:
            cls.querying = True
            t1 = thr(
                target=cls.__action,
                args=(),
            )
            t1.start()
        except Exception as e:
            cls.querying = False
            logging.debug(getExc(e))
            t1.join()

    @classmethod
    def stop_spinning(cls):
        """Stop displaying busy-bar"""
        if cls.querying:
            cls.querying = False
            sleep(cls.sleep_time)

    @classmethod
    def run(cls, help: str = "Exception"):
        """Handle function exceptions safely why showing busy bar

        Args:
            help (str, optional): Message to be shown incase of an exception. Defaults to ''.
        """

        def decorator(func):
            @wraps(func)  # Preserves function metadata
            def main(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except KeyboardInterrupt:
                    cls.stop_spinning()
                    return
                except EOFError:
                    cls.querying = False
                    exit(logging.info("Stopping program"))
                except Exception as e:
                    cls.stop_spinning()
                    logging.error(f"{help} - {getExc(e)}")

            return main

        return decorator


class InteractiveChatGPT(cmd.Cmd):
    intro = f"Welcome to {__info__} Type help <command> or h for general help info."
    prompt = (
        f"┌─[{getpass.getuser().capitalize()}@chatgpt-anon](v{__version__})\r\n└──╼ ❯❯❯"
    )

    def __init__(self, model, timeout, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.timeout = timeout
        self.bot = ChatGPT(model=model, timeout=timeout)
        self.user_name = getpass.getuser().capitalize()
        self.prettify = True
        self.color = "cyan"
        self.show_title = False
        self.code_theme = "monokai"
        self.quiet = False
        self.vertical_overflow = "ellipsis"

    def output_bond(
        self,
        title: str,
        text: str,
        color: str = "cyan",
        frame: bool = True,
        is_json: bool = False,
    ):
        """Print prettified output

        Args:
            title (str): Title
            text (str): Info to be printed
            color (str, optional): Output color. Defaults to "cyan".
            frame (bool, optional): Add frame. Defaults to True.
        """
        if is_json:
            text = f"""
```json
{json.dumps(text,indent=4)}
```
"""
        rich.print(
            Panel(
                Markdown(text, code_theme=self.code_theme),
                title=title.title(),
                style=Style(
                    color=color,
                    frame=frame,
                ),
            ),
        )
        if is_json and click.confirm("Do you wish to save this"):
            default_path = title + ".json"
            save_to = click.prompt(
                "Enter path to save to", default=default_path, type=click.STRING
            )
            with open(save_to, "a") as fh:
                json.dump(text, fh, indent=4)
            click.secho(f"Successfuly saved to `{save_to}`", fg="green")

    def do_h(self, line):
        """Show help info in tabular format"""
        table = Table(
            title="Help info",
            show_lines=True,
        )
        table.add_column("No.", style="white", justify="center")
        table.add_column("Command", style="yellow", justify="left")
        table.add_column("Function", style="cyan")
        command_methods = [
            getattr(self, method)
            for method in dir(self)
            if callable(getattr(self, method)) and method.startswith("do_")
        ]
        command_methods.append(self.default)
        command_methods.reverse()
        for no, method in enumerate(command_methods):
            table.add_row(
                str(no + 1),
                method.__name__[3:] if not method == self.default else method.__name__,
                method.__doc__,
            )
        Console().print(table)
        click.secho(f"Submit any bug at : {__repo__}/issues/new", fg="yellow")

    @busy_bar.run(help="Ensure conversation ID is correct")
    def do_history(self, line):
        """Show conversation history"""
        history = self.bot.chat_history(
            conversation_id=click.prompt(
                "Conversation ID",
                default=self.bot.current_conversation_id,
                type=click.STRING,
            )
        )
        formatted_chats = []
        format_datetime = (
            lambda timestamp: datetime.datetime.fromtimestamp(timestamp)
            .today()
            .strftime("%H:%M:%S %d-%b-%Y")
        )

        for entry in history.get("content"):
            formatted_chats.append(
                f"""
### {entry['author']} (**{format_datetime(entry['create_time'])}**)

{entry['text']}
"""
            )
        self.output_bond(
            history.get("title"),
            "\n\n".join(formatted_chats),
        )
        if click.confirm("Do you wish to save this"):
            path = click.prompt(
                "Enter path to save to",
                default=history.get("title") + ".json",
                type=click.STRING,
            )
            with open(path, "a") as fh:
                json.dump(
                    history,
                    fh,
                    indent=click.prompt(
                        "Json Indentantion level",
                        default=4,
                        type=click.INT,
                    ),
                )
            click.secho(f"Saved successfully to `{path}`")

    @busy_bar.run(help="Ensure conversation ID is correct")
    def do_share(self, line):
        """Share conversation by link"""
        share_info = self.bot.share_conversation(
            conversation_id=click.prompt(
                "Conversation ID",
                default=self.bot.current_conversation_id,
                type=click.STRING,
            ),
            is_anonymous=click.confirm("Is anonymous", default=True),
            is_public=click.confirm("Is public", default=True),
            is_visible=True,
        )
        url = share_info.get("share_url")
        self.output_bond(share_info.get("title"), f"Url : **{url}**")
        if click.confirm("Copy link to clipboard"):
            clipman.set(url)
            click.secho("Link copied to clipboard.", fg="green")

    @busy_bar.run(help="Probably conversation ID is incorrect")
    def do_stop_share(self, line):
        """Revoke shared conversation link"""
        success_report = self.bot.stop_sharing_conversation(
            self.bot.share_conversation(
                conversation_id=click.prompt(
                    "Conversation ID",
                    default=self.bot.current_conversation_id,
                    type=click.STRING,
                ),
            ).get("share_id")
        )
        self.output_bond("Success Report", success_report, is_json=True)

    @busy_bar.run(help="Probably conversation ID is incorrect")
    def do_rename(self, line):
        """Rename conversation title"""
        new_title = click.prompt("New title", default=line, type=click.STRING)
        if click.confirm("Are you sure to change conversation title"):
            response = self.bot.rename_conversation(
                conversation_id=click.prompt(
                    "Conversation ID",
                    default=self.bot.current_conversation_id,
                    type=click.STRING,
                ),
                title=new_title,
            )
            self.output_bond("Change Convo Title", response, is_json=True)
        else:
            click.secho("Conversation title retained", fg="yellow")

    @busy_bar.run(help="Probably conversation ID is incorrect")
    def do_archive(self, line):
        """Archive or unarchive a conversation"""
        conversation_id = click.prompt(
            "Conversation ID",
            default=self.bot.current_conversation_id,
            type=click.STRING,
        )
        is_archive = click.confirm(
            "Is archive",
            default=True,
        )
        if click.confirm("Are you sure to perform this operation"):
            response = self.bot.archive_conversation(
                conversation_id,
                is_archived=is_archive,
            )
            self.output_bond("Archive Report", response, is_json=True)

    @busy_bar.run()
    def do_shared_conversations(self, line):
        """Show shared conversations"""
        shared = self.bot.shared_conversations()
        self.output_bond("Shared Conversations", shared, is_json=True)

    @busy_bar.run()
    def do_previous_conversations(self, line):
        """Show previous conversations"""
        previous_convos = self.bot.previous_conversations(
            limit=click.prompt("Convesation limit", type=click.INT, default=28),
            offset=click.prompt("Conversation offset", type=click.INT, default=0),
            all=True,
        )
        self.output_bond("Previous Conversations", previous_convos, is_json=True)

    @busy_bar.run(help="Probably conversation ID is incorrect")
    def do_delete_conversation(self, line):
        """Delete a particular conversation"""
        conversation_id = click.prompt(
            "Conversation ID",
            default=self.bot.current_conversation_id,
            type=click.STRING,
        )
        if click.confirm("Are you sure to delete this conversation"):
            response = self.bot.delete_conversation(
                conversation_id,
            )
            self.output_bond("Deletion Report", response, is_json=True)

    @busy_bar.run()
    def do_prompts(self, line):
        """Generate random prompts"""
        prompts = self.bot.prompt_library(
            limit=click.prompt("Total prompts", type=click.INT, default=4),
        )
        self.output_bond("Random Prompts", prompts, is_json=True)

    @busy_bar.run()
    def do_account_info(self, line):
        """Show information related to current account at ChatGPT"""
        details = self.bot.user_details(
            in_details=click.confirm("Show in details", default=True),
        )
        self.output_bond("Account Info", details, is_json=True)

    @busy_bar.run()
    def do_ask(self, line):
        """Show raw response from ChatGPT"""
        response = self.bot.ask(
            prompt=(
                line
                if bool(line.strip())
                else click.prompt("Prompt", type=click.STRING)
            )
        )
        self.output_bond("Raw Response", response, is_json=True)

    @busy_bar.run()
    def do_auth(self, line):
        """Show current user auth info"""
        if click.confirm(
            "Contents to be displayed contains sensitive data. Are you sure to continue",
        ):
            self.output_bond("Current Auth info", self.bot.auth, is_json=True)

    @busy_bar.run()
    def do_migrate(self, line):
        """Shift to another conversation"""
        if click.confirm(
            "Are you sure to shift to new conversation",
        ):
            self.model = click.prompt(
                "ChatGPT model", default=self.model, type=click.STRING
            )
            self.timeout = click.prompt(
                "Request timeout",
                default=self.timeout,
                type=click.INT,
            )
            self.bot = ChatGPT(
                model=self.model,
                timeout=self.timeout,
            )

    @busy_bar.run()
    def do_exit(self, line):
        """Quit this program"""
        if click.confirm("Are you sure to exit"):
            print("Okay Goodbye!")
            return True

    @busy_bar.run()
    def do_set_theme(self, line):
        """Set theme for displaying codes"""
        if line in rich_code_themes:
            self.code_theme = line
        else:
            self.code_theme = Prompt.ask(
                "Enter theme name",
                choices=rich_code_themes,
            )
        click.secho(f"Code theme set to '{self.code_theme}'")

    def generate_title(self):
        """Get current conversation title"""
        resp = self.bot.generate_title(
            self.bot.current_conversation_id,
            self.bot.get_current_message_id(),
        )
        if "message" in resp:
            return resp["message"]
        return "Untitled"

    @busy_bar.run()
    def do_with_copied(self, line):
        """Attach last copied text to the prompt
        Usage:
            from_copied:
                 prompt = {text-copied}
            from_copied Debug this code:
                 prompt = Debug this code {newline} {text-copied}
        """
        issued_prompt = (
            f"{line}\n{clipman.get()}" if bool(line.strip()) else clipman.get()
        )
        click.secho(issued_prompt, fg="yellow")
        if click.confirm("Do you wish to proceed"):
            self.default(issued_prompt)

    @busy_bar.run(help="System error")
    def do_copy_this(self, line):
        """Copy last response
        Usage:
           copy_this:
               text-copied = {whole last-response}
           copy_this code:
               text-copied = {All codes in last response}
        """
        if self.bot.last_response:
            global last_response
            last_response = get_message(self.bot.last_response)
            if not "code" in line:
                clipman.set(last_response)
                click.secho("Last response copied successfully!", fg="cyan")
                return

            # Copies just code
            sanitized_codes = []
            code_blocks = re.findall(r"```.*?```", last_response, re.DOTALL)
            for code_block in code_blocks:
                new_code_block = re.sub(
                    "^```.*$", "", code_block.strip(), flags=re.MULTILINE
                )
                if bool(new_code_block.strip()):
                    sanitized_codes.append(new_code_block)
            if sanitized_codes:
                if len(sanitized_codes) > 1:
                    if not click.confirm("Do you wish to copy all codes"):
                        for index, code in enumerate(sanitized_codes):
                            rich.print(
                                Panel(
                                    Markdown(
                                        code_blocks[index], code_theme=self.code_theme
                                    ),
                                    title=f"Index : {index}",
                                    title_align="left",
                                )
                            )

                        clipman.set(
                            sanitized_codes[
                                click.prompt(
                                    "Enter code index",
                                    type=click.IntRange(0, len(sanitized_codes) - 1),
                                )
                            ]
                        )
                        click.secho("Code copied successfully", fg="cyan")
                    else:
                        clipman.set("\n\n".join(sanitized_codes))
                        click.secho(
                            f"All {len(sanitized_codes)} codes copied successfully!",
                            fg="cyan",
                        )
                else:
                    clipman.set(sanitized_codes[0])
                    click.secho("Code copied successfully!", fg="cyan")
            else:
                click.secho("No code found in the last response!", fg="red")
        else:
            click.secho("Chat with ChatGPT first.", fg="yellow")

    def do_clear(self, line):
        """Clear console"""
        sys.stdout.write("\u001b[2J\u001b[H")
        sys.stdout.flush()

    # @busy_bar.run()
    def default(self, line):
        """Chat with ChatGPT"""
        if line.startswith("./"):
            os.system(line[2:])
        else:
            try:
                busy_bar.start_spinning()
                generated_response = self.bot.chat(line, stream=True)
                busy_bar.stop_spinning()
                stdout_handler = stream_console_output if self.quiet else stream_output
                stdout_handler(
                    generated_response,
                    title="",
                    is_markdown=self.prettify,
                    style=Style(
                        color=self.color,
                    ),
                    title_generator=self.generate_title if self.show_title else None,
                    code_theme=self.code_theme,
                    vertical_overflow=self.vertical_overflow,
                )
                """
                if self.prettify:
                    rich.print(Markdown(generated_response))
                else:
                    click.secho(generated_response)
                """

            except (KeyboardInterrupt, EOFError):
                busy_bar.stop_spinning()
                print("")
                return False  # Exit cmd

            except Exception as e:
                # logging.exception(e)
                busy_bar.stop_spinning()
                logging.error(getExc(e))


@click.group("chat")
@click.version_option(__version__, "-v", "--version", package_name="chatgpt-anon")
@click.help_option("-h", "--help")
def chat():
    """Reverse Engineered ChatGPT Web-Version"""
    pass


@chat.command()
@click.option(
    "-M",
    "--model",
    help="ChatGPT's model to be used",
    envvar="chatgpt_model",
    default="text-davinci-002-render-sha",
)
@click.option(
    "-T",
    "--timeout",
    help="Http request timeout",
    type=click.INT,
    default=30,
)
@click.argument(
    "prompt",
    required=False,
)
@click.option(
    "-B",
    "--busy-bar-index",
    help="Busy bar index [0: None, 1:/, 2:■█■■■, 3:⣻]",
    type=click.IntRange(0, 3),
    default=1,
    envvar="busy_bar_index",
)
@click.option(
    "-T",
    "--code-theme",
    help="Theme for styling codes in markdown",
    default="monokai",
    type=click.Choice(rich_code_themes),
)
@click.option(
    "-c", "--color", default=None, help="Font color for printing the contents"
)
@click.option(
    "-vo",
    "--vertical-overflow",
    envvar="vertical_overflow",
    help="Vertical overflow behaviour on content display",
    type=click.Choice(["visible", "crop", "ellipsis"]),
    default="ellipsis",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    default=False,
    help="Disable response framing - Defaults to False",
    envvar="quiet",
)
@click.option("--prettify/--raw", default=True, help="Prettify the markdowned response")
@click.option(
    "--show-title/--no-title", default=True, help="Flag for title generation control"
)
@click.help_option("-h", "--help")
def interactive(
    model,
    timeout,
    prompt,
    busy_bar_index,
    code_theme,
    color,
    vertical_overflow,
    quiet,
    prettify,
    show_title,
):
    """Chat with ChatGPT interactively"""
    rich.print(
        Panel(
            f"""
            Repo : {__repo__}
            By : {__author__}
          """,
            title=f"chatgpt-anon v{__version__}",
            style=Style(
                color="cyan",
                frame=True,
            ),
        ),
    )
    busy_bar.spin_index = busy_bar_index
    bot = InteractiveChatGPT(model, timeout)
    bot.prettify = prettify
    bot.color = color
    bot.show_title = show_title
    bot.code_theme = code_theme
    bot.quiet = quiet
    bot.vertical_overflow = vertical_overflow
    if prompt:
        bot.default(prompt)
    bot.cmdloop()


@chat.command()
@click.option(
    "-M",
    "--model",
    help="ChatGPT's model to be used",
    envvar="chatgpt_model",
    default="text-davinci-002-render-sha",
)
@click.option(
    "-T",
    "--timeout",
    help="Http request timeout",
    type=click.INT,
    default=30,
)
@click.argument("prompt", required=True)
@click.option(
    "-T",
    "--code-theme",
    help="Theme for styling codes in markdown",
    default="monokai",
    type=click.Choice(rich_code_themes),
)
@click.option(
    "-c", "--color", default=None, help="Font color for printing the contents"
)
@click.option(
    "-B",
    "--busy-bar-index",
    help="Busy bar index [0: None, 1:/, 2:■█■■■, 3:⣻]",
    type=click.IntRange(0, 3),
    default=1,
    envvar="busy_bar_index",
)
@click.option(
    "-vo",
    "--vertical-overflow",
    envvar="vertical_overflow",
    help="Vertical overflow behaviour on content display",
    type=click.Choice(["visible", "crop", "ellipsis"]),
    default="ellipsis",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    default=False,
    help="Disable response framing - Defaults to False",
    envvar="quiet",
)
@click.option("--prettify/--raw", default=True, help="Prettify the markdowned response")
@click.help_option("-h", "--help")
def generate(
    model,
    timeout,
    prompt,
    code_theme,
    color,
    busy_bar_index,
    vertical_overflow,
    quiet,
    prettify,
):
    """Generate a quick response with ChatGPT (Default)"""

    # bot = ChatGPT(cookie_path, model, index, timeout=timeout)
    busy_bar.spin_index = busy_bar_index
    bot = InteractiveChatGPT(model, timeout)
    bot.prettify = prettify
    bot.color = color
    bot.code_theme = code_theme
    bot.quiet = quiet
    bot.vertical_overflow = vertical_overflow
    bot.default(prompt)


@error_handler(exit_on_error=True)
def main(*cmd_args):
    sys.argv += list(cmd_args)
    dotenv.load_dotenv(os.path.join(os.getcwd(), ".env"))
    args = sys.argv
    if len(args) > 1 and args[1] not in chat.commands.keys() and not "-" in args[1]:
        # Just an hack to make 'generate' default option
        sys.argv.insert(1, "generate")
    chat()



#!/usr/bin/python
import requests
from chatgpt_anon import utils
import json
import re
from functools import lru_cache
from typing import Iterator


class ChatGPT:
    def __init__(
        self,
        model: str = "text-davinci-002-render-sha",
        locale: str = "en-US",
        user_agent: str = utils.request_headers["User-Agent"],
        timeout: tuple = 30,
        disable_history_and_training: bool = False,
    ):
        """Initializes ChatGPT

        Args:
            cookie_path (str): Path to `.json` file containing `chat.openai.com` cookies
            model (str, optional): ChatGPT text generation model name. Defaults to `text-davinci-002-render-sha`.
            conversation_index (int, optional): Conversation index to pick up conversation from. Defaults to `1`.
            locale (str, optional): Your locale. Defaults to `en-US`
            user_agent (str, optional): Http request header User-Agent. Defaults to `Mozilla/5.0 (X11; Linux x86_64; rv:120.0) Gecko/20100101 Firefox/120.0`
            timeout (int, optional): Http request timeout.
            trace (bool, optional): Trace websocket requests. Defaults to False.

        """
        self.session = requests.Session()
        self.timeout = timeout
        self.auth = {}  # Will be updated while updating headers
        # self.session.cookies.update(utils.get_cookies(cookie_path))
        self.session.headers.update(
            utils.get_request_headers_and_initialize_request(self)
        )
        self.conversation_endpoint = "https://chat.openai.com/backend-anon/conversation"
        self.account_detail_endpoint = (
            "https://chat.openai.com/backend-anon/accounts/check"
        )
        self.account_details_endpoint = (
            self.account_detail_endpoint + "/v4-2023-04-27"
        )  # update the date as frequent as possible
        self.prompt_library_endpoint = (
            "https://chat.openai.com/backend-anon/prompt_library/"
        )
        self.previous_conversations_endpoint = (
            "https://chat.openai.com/backend-anon/conversations"
        )
        self.title_generation_endpoint = "https://chat.openai.com/backend-anon/conversation/gen_title/%(conversation_id)s"
        self.conversation_manipulation_endpoint = (
            "https://chat.openai.com/backend-anon/conversation/%(conversation_id)s"
        )
        self.share_conversation_endpoint = (
            "https://chat.openai.com/backend-anon/share/create"
        )
        self.share_conversation_patch_endpoint = (
            "https://chat.openai.com/backend-anon/share/%(share_id)s"
        )
        self.shared_conversations_endpoint = (
            "https://chat.openai.com/backend-anon/shared_conversations"
        )
        self.shared_conversation_view_endpoint = (
            "https://chat.openai.com/share/%(share_id)s"
        )
        self.stop_sharing_conversation_endpoint = (
            "https://chat.openai.com/backend-anon/%(share_id)s"
        )
        self.sentinel_chat_requirements_endpoint: str = (
            "https://chat.openai.com/backend-anon/sentinel/chat-requirements"
        )
        self.session.headers["User-Agent"] = user_agent
        self.locale = locale
        self.model = model
        self.disable_history_and_training = disable_history_and_training
        self.last_response = {}
        self.last_response_metadata = {
            1: {},
            2: {},
        }
        self.__already_init = False
        self.__title_cache = {}
        self.stream_chunk_size = 64

    def __generate_payload(self, prompt: str) -> dict:
        return utils.generate_payload(self, prompt)

    @property
    def current_conversation_id(self):
        if self.__already_init:
            return self.last_response_metadata.get(2).get("conversation_id")
        else:
            self.__already_init = True
            return None

    def get_current_message_id(self):
        return self.last_response_metadata.get(2).get("message_id")

    def update_sentinel_tokens(self):
        resp = self.session.post(self.sentinel_chat_requirements_endpoint, json={})
        resp.raise_for_status()
        self.session.headers.update(
            {"OpenAI-Sentinel-Chat-Requirements-Token": resp.json()["token"]}
        )

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw_response: bool = False,
    ) -> dict | Iterator:
        """Chat with ChatGPT

                Args:
                    prompt (str): message to be send
                    stream (bool, optional): Flag to stream response. Defaults to False.
                    raw_response (bool, optional): Yield back unmodified response
                returns :
                    dict {}
                Yields :
                   Iterator[dict]
        ```json
        {
            "message": {
                "id": "2c98d9ff-495c-4f08-af9e-affbd17xxxxx",
                "author": {
                    "role": "assistant",
                    "name": null,
                    "metadata": {}
                },
                "create_time": 1702666802.823688,
                "update_time": null,
                "content": {
                    "content_type": "text",
                    "parts": [
                        "Of course, your privacy matters! I don't store or remember our conversations once they're completed, so your information is kept confidential. If there's anything specific you'd like to discuss or if you have any concerns, feel free to let me know. I'm here to assist you!"
                    ]
                },
                "status": "finished_successfully",
                "end_turn": true,
                "weight": 1.0,
                "metadata": {
                    "finish_details": {
                        "type": "stop",
                        "stop_tokens": [
                            100260
                        ]
                    },
                    "inline_gizmo_id": null,
                    "is_complete": true,
                    "message_type": "next",
                    "model_slug": "text-davinci-002-render-sha",
                    "parent_id": "7bf27013-a47a-438c-ae17-0ee846b4xxxx",
                    "timestamp_": "absolute"
                },
                "recipient": "all"
            },
            "conversation_id": "affdda8c-588c-4342-9869-26c5bd7xxxxx",
            "error": null
        }
        ```
        """
        payload: dict = self.__generate_payload(prompt)
        # print(json.dumps(payload, indent=4))
        self.update_sentinel_tokens()
        response = self.session.post(
            url=self.conversation_endpoint,
            json=payload,
            timeout=self.timeout,
            stream=True,
        )
        # response.raise_for_status()
        if (
            response.ok
            and response.headers.get("content-type")
            == "text/event-stream; charset=utf-8"
        ):

            def for_stream():
                for value in response.iter_lines(
                    decode_unicode=True,
                    delimiter="data:",
                    chunk_size=self.stream_chunk_size,
                ):
                    try:
                        to_dict = json.loads(value)
                        if "is_completion" in to_dict.keys():
                            # Metadata (response)
                            self.last_response_metadata[
                                2 if to_dict.get("is_completion") else 1
                            ] = to_dict
                            continue
                        # Only data containing the `feedback body` make it to here
                        self.last_response.update(to_dict)
                        yield value if raw_response else to_dict
                    except json.decoder.JSONDecodeError:
                        # Caused by either empty string or [DONE]
                        if raw_response:
                            yield value
                        pass

            def for_non_stream():
                response_to_be_returned = {}
                for value in response.iter_lines(
                    decode_unicode=True,
                    delimiter="data:",
                    chunk_size=self.stream_chunk_size,
                ):
                    try:
                        to_dict = json.loads(value)
                        if "is_completion" in to_dict.keys():
                            # Metadata (response)
                            self.last_response_metadata[
                                2 if to_dict.get("is_completion") else 1
                            ] = to_dict
                            continue
                        # Only data containing the `feedback body` make it to here
                        self.last_response.update(to_dict)
                        response_to_be_returned.update(to_dict)
                    except json.decoder.JSONDecodeError:
                        # Caused by either empty string or [DONE]
                        pass
                return response_to_be_returned

            return for_stream() if stream else for_non_stream()

        else:
            raise Exception(
                f"Failed to fetch response - ({response.status_code}, {response.reason} : {response.headers.get('content-type')} : {response.text})"
            )

    def chat(self, prompt: str, stream: bool = False) -> str:
        """Interact with ChatGPT on the fly

        Args:
            prompt (str): Message to ChatGPT
            stream (bool, optional): Yield the text response. Defaults to False.

        Returns:
            str: Text response generated only

        Yields:
            Iterator[str]: Text response generated only
        """
        resp = self.ask(prompt, stream)

        def for_stream():
            # streaming response
            for response in resp:
                yield utils.get_message(response)
                # pass  Currently fixed

        def for_non_stream():
            return utils.get_message(resp)

        return for_stream() if stream else for_non_stream()

    def user_details(self, in_details: bool = True) -> dict:
        """Returns various information concerning the user

                Args:
                    in_details (bool, optional): Return detailed info. Defaults to True.
                returns:
                    dict : {}
        ```json
        {
             "accounts": {
                  "b8a156d7-9a30-4de3-bf40-4d88782xxxxx": {
                       "account": {
                            "account_user_role": "account-owner",
                            "account_user_id": "user-IUW2uKHQ6bwv4bdiI8IYxxxx__b8a156d7-9a30-4de3-bf40-4d887829xxxx",
                            "processor": {
                                 "a001": {
                                      "has_customer_object": false
                                 },
                                 "b001": {
                                      "has_transaction_history": false
                                 },
                                 "c001": {
                                      "has_transaction_history": false
                                 }
                            },
                            "account_id": "b8a156d7-xxxx-4de3-bf40-4d887829xxxx",
                            "organization_id": null,
                            "is_most_recent_expired_subscription_gratis": false,
                            "has_previously_paid_subscription": false,
                            "name": null,
                            "profile_picture_id": null,
                            "profile_picture_url": null,
                            "structure": "personal",
                            "plan_type": "free",
                            "is_deactivated": false,
                            "promo_data": {}
                       },
                       "features": [
                            "allow_url_thread_creation",
                            "arkose_enabled",
                            "arkose_gpt_35_experiment",
                            "bizmo_settings",
                            "breeze_available",
                            "chat_preferences_available",
                            "conversation_bot_arkose",
                            "disable_team_upgrade_ui",
                            "gizmo_live",
                            "gizmo_ui",
                            "invite_referral",
                            "new_plugin_oauth_endpoint",
                            "privacy_policy_nov_2023",
                            "shareable_links",
                            "starter_prompts",
                            "user_settings_announcements"
                       ],
                       "entitlement": {
                            "subscription_id": null,
                            "has_active_subscription": false,
                            "subscription_plan": "chatgptfreeplan",
                            "expires_at": null
                       },
                       "last_active_subscription": {
                            "subscription_id": null,
                            "purchase_origin_platform": "chatgpt_not_purchased",
                            "will_renew": false
                       }
                  },
                  "default": {
                       "account": {
                            "account_user_role": "account-owner",
                            "account_user_id": "user-IUW2uKHQ6bwv4bdiI8IYtjqA__b8a156d7-9a30-4de3-bf40-xxxxxxxxxx",
                            "processor": {
                                 "a001": {
                                      "has_customer_object": false
                                 },
                                 "b001": {
                                      "has_transaction_history": false
                                 },
                                 "c001": {
                                      "has_transaction_history": false
                                 }
                            },
                            "account_id": "b8a156d7-9a30-4de3-bf40-4d887829xxxx",
                            "organization_id": null,
                            "is_most_recent_expired_subscription_gratis": false,
                            "has_previously_paid_subscription": false,
                            "name": null,
                            "profile_picture_id": null,
                            "profile_picture_url": null,
                            "structure": "personal",
                            "plan_type": "free",
                            "is_deactivated": false,
                            "promo_data": {}
                       },
                       "features": [
                            "allow_url_thread_creation",
                            "arkose_enabled",
                            "arkose_gpt_35_experiment",
                            "bizmo_settings",
                            "breeze_available",
                            "chat_preferences_available",
                            "conversation_bot_arkose",
                            "disable_team_upgrade_ui",
                            "gizmo_live",
                            "gizmo_ui",
                            "invite_referral",
                            "new_plugin_oauth_endpoint",
                            "privacy_policy_nov_2023",
                            "shareable_links",
                            "starter_prompts",
                            "user_settings_announcements"
                       ],
                       "entitlement": {
                            "subscription_id": null,
                            "has_active_subscription": false,
                            "subscription_plan": "chatgptfreeplan",
                            "expires_at": null
                       },
                       "last_active_subscription": {
                            "subscription_id": null,
                            "purchase_origin_platform": "chatgpt_not_purchased",
                            "will_renew": false
                       }
                  }
             },
             "account_ordering": [
                  "b8a156d7-9a30-4de3-bf40-4d88782xxxx"
             ]
        }
        ```
        """
        resp = self.session.get(
            (
                self.account_details_endpoint
                if in_details
                else self.account_detail_endpoint
            ),
            timeout=self.timeout,
        )
        return utils.is_json(resp, "account data")

    def prompt_library(self, limit: int = 4, offset: int = 0) -> list:
        """Generates random prompts

                Args:
                    limit (int, optional): Limit suggestions. Defaults to 4.
                    offset (int, optional): Offset. Defaults to 0.
                returns:
                    list : []

        ```json
           {
            "items": [
                {
                    "id": "edb56xxx",
                    "title": "Design a database schema",
                    "description": "for an online merch store",
                    "prompt": "Design a database schema for an online merch store."
                },
                {
                    "id": "9fa37xxx",
                    "title": "Recommend a dish",
                    "description": "to impress a date who's a picky eater",
                    "prompt": "I'm going to cook for my date who claims to be a picky eater. Can you recommend me a dish that's easy to cook?"
                }
            ],
            "total": 2,
            "limit": 2,
            "offset": 1
            }
        ```
        """

        resp = self.session.get(
            self.prompt_library_endpoint,
            params={"limit": limit, "offset": offset},
            timeout=self.timeout,
        )
        return utils.is_json(resp, "prompts")

    def previous_conversations(
        self,
        limit: int = 20,
        offset: int = 0,
        order: str = "updated",
        index: int = 1,
        all: bool = False,
    ) -> list:
        """Loads previous conversations

                Args:
                    limit (int, optional): Fetch this specific amount of chats. Defaults to 28.
                    offset (int, optional): ``. Defaults to 0.
                    order (str, optional): Sort order. Defaults to "updated".
                    index (int, optional): Index of the item to be returned +1. Defaults to 1.
                    all (bool, optional): Return all conversations based on specified limit

                Returns:
                    list: Previous conversations contained in dict

                ```json
                {
             "items": [
                  {
                       "id": "86f73b54-0f51-47ba-84a3-07c1e25xxxx",
                       "title": "Urine Cleaning in Rome",
                       "create_time": "2023-12-15T13:39:24.683876+00:00",
                       "update_time": "2023-12-15T14:02:23.776574+00:00",
                       "mapping": null,
                       "current_node": null,
                       "conversation_template_id": null,
                       "gizmo_id": null,
                       "is_archived": false,
                       "workspace_id": null
                  },
                  {
                       "id": "b3779121-8767-4202-9527-3058f40xxxx",
                       "title": "Helpful User, Assistant",
                       "create_time": "2023-12-15T12:03:30.596706+00:00",
                       "update_time": "2023-12-15T13:29:05.286457+00:00",
                       "mapping": null,
                       "current_node": null,
                       "conversation_template_id": null,
                       "gizmo_id": null,
                       "is_archived": false,
                       "workspace_id": null
                  }
             ],
             "total": 108,
             "limit": 2,
             "offset": 0,
             "has_missing_conversations": false
           }
        ```
        """
        assert isinstance(index, int), "Index must be an integer"
        index -= 1  # So that 0 equates to False as in self.current_conversation_id
        resp = self.session.get(
            self.previous_conversations_endpoint,
            params={"limit": limit, "offset": offset, "order": order},
            timeout=self.timeout,
        )
        resp = utils.is_json(resp, "conversation history")
        if all:
            return resp
        conversations = resp["items"]
        # conversations.reverse()
        if len(conversations) - 1 >= index:
            return conversations[index]
        else:
            raise Exception(
                f"Index '{index} is greater than the total conversations '{len(conversations)}"
            )

    @lru_cache(maxsize=201)
    def generate_title(
        self, conversation_id: str, message_id: str, per_conversation: bool = True
    ) -> dict:
        """Generates title for a particular conversation message

        Args:
            conversation_id (str): ``
            message_id (str): ``
            per_conversation (bool): Ignore message_id and serve from cache based on conversation_id. Defaults to True.

        Returns:
            dict: `{}`
        """
        if per_conversation and self.__title_cache.get(conversation_id):
            # OpenAI dropped support for generating conversation title for every new message in ChatGPT 3.5 model
            #  - Probably to non-premium users only.
            return self.__title_cache[conversation_id]
        resp = self.session.post(
            self.title_generation_endpoint % {"conversation_id": conversation_id},
            json={"message_id": message_id},
            timeout=self.timeout,
        )
        sanitized_resp = utils.is_json(resp, "title")
        if "message" in sanitized_resp.keys():
            generated_title = sanitized_resp.get("message")
            match = re.search(r"'([^']*)'", generated_title)
            if match:
                sanitized_resp["message"] = match.group(1)
                if per_conversation:
                    self.__title_cache[conversation_id] = sanitized_resp
        return sanitized_resp

    def delete_conversation(self, conversation_id: str) -> dict:
        """Deletes a particular conversation based on ID

        Args:
            conversation_id (str): Conversation iD

        Returns:
            dict: Response
        """
        resp = self.session.patch(
            self.conversation_manipulation_endpoint
            % {"conversation_id": conversation_id},
            json={"is_visible": False},
        )
        return utils.is_json(resp, "delete")

    def chat_history(self, conversation_id: str, all: bool = False) -> dict:
        """Fetches previous chat prompts and responses

                Args:
                    conversation_id (str): Conversation ID
                    all (bool): Return all response as received. Defaults to False.

                Returns:
                    dict: Previous chats

                ```json
                {
            "title": "Trump's Age Calculator",
            "create_time": 1703074882.684634,
            "update_time": 1703074885.46044,
            "current_node": "f18a446d-8843-4433-acf7-79cb01a8xxxx",
            "conversation_id": "00565704-a7ae-4278-bd14-ca598fedxxxx",
            "is_archived": false,
            "moderation_results": [],
            "safe_urls": [],
            "content": [
                {
                    "author": "User",
                    "create_time": 1703074882.685243,
                    "text": "How old is Donald Triump",
                    "status": "finished_successfully",
                    "id": "aaa2ea97-d8bd-4b93-bff9-09d1a684xxxx"
                },
                {
                    "author": "ChatGPT",
                    "create_time": 1703074884.019059,
                    "text": "Donald Trump was born on June 14, 1946, so his age depends on the current date. If you tell me today's date, I can calculate his age for you!",
                    "status": "finished_successfully",
                    "id": "f18a446d-8843-4433-acf7-79cb01a8xxxx"
                }
            ]
        }
                ```
        """
        resp = self.session.get(
            self.conversation_manipulation_endpoint
            % {"conversation_id": conversation_id}
        )
        from_chatgpt = utils.is_json(resp, "chat history")
        if all:
            return from_chatgpt
        # title
        # create_time
        # update_time
        # mapping >
        # <Drop first two>
        # mapping > message_id > message > content > parts[0]
        #                                > create_time
        #                                > status
        new_resp = {
            "title": from_chatgpt["title"],
            "create_time": from_chatgpt["create_time"],
            "update_time": from_chatgpt["update_time"],
            "current_node": from_chatgpt["current_node"],
            "conversation_id": from_chatgpt["conversation_id"],
            "is_archived": from_chatgpt["is_archived"],
            "moderation_results": from_chatgpt["moderation_results"],
            "safe_urls": from_chatgpt["safe_urls"],
            "content": [],
        }
        for count, entry in enumerate(list(from_chatgpt["mapping"].keys())[2:]):
            in_need = from_chatgpt["mapping"][entry]["message"]
            new_resp["content"].append(
                {
                    "author": "ChatGPT" if count % 2 else "User",
                    "create_time": in_need["create_time"],
                    "text": in_need["content"]["parts"][0],
                    "status": in_need["status"],
                    "id": entry,
                }
            )
        return new_resp

    def rename_conversation(self, conversation_id: str, title: str) -> dict:
        """Renames conversation title

                Args:
                    conversation_id (str): Conversation ID
                    title (str): New conversation title

                Returns:
                    dict: Success report
                ```json
                {
            "success": true
            }
        ```
        """
        resp = self.session.patch(
            self.conversation_manipulation_endpoint
            % {"conversation_id": conversation_id},
            json={"title": title},
        )
        return utils.is_json(resp, "rename conversation")

    def archive_conversation(
        self, conversation_id: str, is_archived: bool = True
    ) -> dict:
        """Archives a particular conversation

                Args:
                    conversation_id (str): Conversation ID
                    is_archived (bool): Archive (True) or Unarchive (False). Defaults to `True`.
                Returns:
                    dict: Success report
                ```json
                {
            "success": true
         }
        ```
        """
        resp = self.session.patch(
            self.conversation_manipulation_endpoint
            % {
                "conversation_id": conversation_id,
            },
            json={
                "is_archived": is_archived,
            },
        )
        return utils.is_json(resp, "archive conversation")

    def share_conversation(
        self,
        conversation_id: str,
        is_anonymous: bool = True,
        is_public: bool = True,
        is_visible: bool = True,
    ) -> dict:
        """Generate link for sharing conversation

                Args:
                    conversation_id (str): Conversation ID
                    anonymous (bool, optional): Hide your Identity in the share. Defaults to True.

                Returns:
                    dict: Success report
                ```json
                {
            "share_id": "a71119f8-9a49-4c1d-b18f-d698313bxxxx",
            "share_url": "https://chat.openai.com/share/a71119f8-9a49-4c1d-b18f-d698313bxxxx",
            "title": "Trump's Age Calculator",
            "is_public": false,
            "is_visible": true,
            "is_anonymous": true,
            "highlighted_message_id": null,
            "current_node_id": "f18a446d-8843-4433-acf7-79cb01a8xxxx",
            "already_exists": false,
            "moderation_state": {
                "has_been_moderated": false,
                "has_been_blocked": false,
                "has_been_accepted": false,
                "has_been_auto_blocked": false,
                "has_been_auto_moderated": false
            }
        }
                ```
        """
        resp = self.session.post(
            self.share_conversation_endpoint,
            json={
                "conversation_id": conversation_id,
                "current_node_id": self.chat_history(conversation_id, True)[
                    "current_node"
                ],
                "is_anonymous": is_anonymous,
            },
        )
        resp_1 = utils.is_json(resp, "share link")
        resp_2 = self.session.patch(
            self.share_conversation_patch_endpoint % dict(share_id=resp_1["share_id"]),
            json={
                "share_id": resp_1["share_id"],
                "highlighted_message_id": resp_1["highlighted_message_id"],
                "title": resp_1["title"],
                "is_public": is_public,
                "is_visible": is_visible,
                "is_anonymous": is_anonymous,
            },
        )
        resp_1.update(utils.is_json(resp_2, "patch share link"))
        return resp_1

    def shared_conversations(self, order: str = "created") -> dict:
        """Get previously shared conversations

                Args:
                    order (str, optional): Sorting paran. Defaults to 'created'.

                Returns:
                    dict: Conversations shared
                ```json
            {
            "items": [
                {
                    "id": "57cf604b-37d6-4910-a47c-41xxxxxxxxxx",
                    "title": "Obama's Age: 62 Years",
                    "create_time": "2023-12-20T19:34:47.883282+00:00",
                    "update_time": "2023-12-20T19:44:35+00:00",
                    "mapping": null,
                    "current_node": null,
                    "conversation_template_id": null,
                    "gizmo_id": null,
                    "is_archived": null,
                    "workspace_id": null,
                    "conversation_id": "f8968cc4-8a48-4771-b16c-58xxxxxxxxxx",
                    "url": "https://chat.openai.com/share/57cf604b-37d6-4910-a47c-41xxxxxxxxxx"
                }
            ],
            "total": 10,
            "limit": 50,
            "offset": 0,
            "has_missing_conversations": false
        }


        """
        resp = self.session.get(
            self.shared_conversations_endpoint,
        )
        shareds = utils.is_json(resp)
        for index, entry in enumerate(shareds["items"]):
            # appends view url to each conversation
            shareds["items"][index]["url"] = (
                self.shared_conversation_view_endpoint % dict(share_id=entry["id"])
            )
        return shareds

    def stop_sharing_conversation(self, share_id: str) -> dict:
        """Deletes sharing link

        Args:
            share_id (str): Shared conversation ID

        Returns:
            dict: Success rate
        """
        resp = self.session.delete(
            self.stop_sharing_conversation_endpoint % dict(share_id=share_id)
        )
        return utils.is_json(resp, "delete shared conversation")



import unittest
from WebChatGPT import ChatGPT


class TestChatGPT(unittest.TestCase):
    @unittest.expectedFailure
    def SetUp(self):
        self.bot = ChatGPT("", "")

    @unittest.expectedFailure
    def test_bot(self):
        """Testing bot"""
        self.assertIsInstance(self.bot.ask("hello world"), dict)

    @unittest.expectedFailure
    def test_chat(self):
        """Testing chat"""
        self.assertIs(self.bot.chat("hello there"), str)

    @unittest.expectedFailure
    def test_user_details(self):
        self.assertIs(self.bot.user_details(), dict)
        self.assertIs(self.bot.user_details(in_details=True), dict)

    @unittest.expectedFailure
    def test_prompt_library(self):
        self.assertIs(self.bot.prompt_library(), dict)

    @unittest.expectedFailure
    def test_previous_conversations(self):
        self.assertIs(self.bot.previous_conversations(), dict)

    @unittest.expectedFailure
    def test_gen_title(self):
        self.assertIs(self.bot.gen_title("", ""), dict)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime, timezone
import json
import logging
import locale
from uuid import uuid4
from typing import Any
from .errors import CookieExpiredError
from .errors import VerificationError


__common_error_support_info = "Incase you have no idea on how to get the cookies check the documentation at - \
    https://github.com/Simatwa/WebChatGPT"


headers = request_headers = {
    "Accept": "text/event-stream",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,en;q=0.5",
    # "Alt-Used": "chat.openai.com",
    "Connection": "keep-alive",
    "Content-Type": "application/json",
    "Host": "chat.openai.com",
    "Origin": "https://chat.openai.com",
    "Referer": "https://chat.openai.com/",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-origin",
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:124.0) Gecko/20100101 Firefox/124.0",
    "OAI-Language": locale.getlocale()[0].replace("_", "-"),
}

init_headers = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,en;q=0.5",
    "User-Agent": headers["User-Agent"],
    "Connection": "keep-alive",
    "Sec-Fetch-Site": "none",
}

response_example = {
    "message": {
        "id": "b9ce2438-a00a-4c1e-9487-3abba1xxxxx",
        "author": {"role": "assistant", "name": None, "metadata": {}},
        "create_time": 1708178888.142936,
        "update_time": None,
        "content": {
            "content_type": "text",
            "parts": ["This is response generated by chatgpt"],
        },
        "status": "in_progress",
        "end_turn": None,
        "weight": 1.0,
        "metadata": {
            "citations": [],
            "gizmo_id": None,
            "message_type": "next",
            "model_slug": "text-davinci-002-render-sha",
            "parent_id": "13442586-492c-4d0e-95a8-032b6xxxxx",
        },
        "recipient": "all",
    },
    "conversation_id": "f5a118a5-e9b5-476c-ac12-e44ad77exxxx",
    "error": None,
}


def error_handler(
    exit_on_error: bool = False, default: Any = None, info: str = None, raise_err=False
):
    """Decorator for handling exceptions

    Args:
        exit_on_error (bool, optional): ``. Defaults to False.
        default (Any, optional): Return this incase of an exception. Defaults to None.
        info (str, optional): Message to be shown along with the exception details. Defaults to None.
        raise_err (bool, optional): Raise exception instead of logging it. Defaults to False.
    """

    def decorator(func):
        def main(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if raise_err:  # Raise exception
                    if info:
                        raise Exception(info) from e
                    else:
                        raise Exception(e)
                else:
                    logging.error(
                        f"{e.args[1] if len(e.args) > 1 else str(e)} - {info}"
                    )
                if exit_on_error:
                    exit(1)
                return default

        return main

    return decorator  #


def get_request_headers_and_initialize_request(self) -> dict:
    """Generate Http request headers & append OAuth

    Returns:
        dict: Request headers
    """
    resp1 = self.session.get("https://chat.openai.com", headers=init_headers)
    resp1.raise_for_status()
    """
    resp = self.session.get(
        "https://chat.openai.com/api/auth/session",
        headers=request_headers,
    )
    #print(json.dumps(resp.json(), indent=4))
    """
    return headers


@error_handler(
    exit_on_error=True,
    raise_err=True,
    info="" + __common_error_support_info,
)
def get_cookies(path: str) -> dict:
    """Reads cookies and format them

    Args:
        path (str): Path to cookie file

    Returns:
        dict: Cookies sorted {name :  value}
    """
    resp = {}
    with open(path) as fh:
        for entry in json.load(fh):
            resp[entry["name"]] = entry["value"]
        return resp


def is_json(response: object, info: str = ""):
    """Checks whether the response is application/json formatted

    Args:
        response (object): `requests.get/post response`
        info (str): Data being fetched
    rtype : dict
    """
    content_type = response.headers.get("content-type")
    if not "application/json" in content_type:
        raise Exception(
            f"Failed to fetch {info} - ({response.status_code}, {response.reason}) `{content_type}` : \n {response.text}"
        )
    return response.json()


def current_timestamp():
    """Generates current timestamp in UTC
    rtype : str
    """
    current_time = datetime.now(timezone.utc)
    # Format the current time in the desired format with microseconds limited to 3 digits
    return current_time.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


def generate_telemetry_payload(self: object):
    """Generates Telemeries - though not implemented in this script but it's good to be aware of.

    Args:
        self (object): ChatGPT class
    rtype : dict
    """
    headers = self.session.headers
    payload_template = {
        "batch": [
            {
                "timestamp": f"{current_timestamp()}",
                "integrations": {"Segment.io": True},
                "userId": f"{headers.get('ajs_user_id')}",
                "anonymousId": f"{headers.get('ajs_anonymous_id')}",
                "event": "Show Starter Prompts",
                "type": "track",
                "properties": {
                    "prompt_count": 4,
                    "prompt_type": "starter",
                    "titles": [entry["title"] for entry in self.prompt_library()],
                    "origin": "chat",
                    "openai_app": "API",
                },
                "context": {
                    "page": {
                        "path": "/",
                        "referrer": "",
                        "search": "",
                        "title": "ChatGPT",
                        "url": "https://chat.openai.com/",
                    },
                    "userAgent": self.session.headers["User-Agent"],
                    "locale": self.locale,
                    "library": {"name": "analytics.js", "version": "npm:next-1.56.0"},
                },
                # "messageId": "ajs-next-3154852a6626ae6a48a031e2506fexxx",
                "_metadata": {
                    "bundled": ["Segment.io"],
                    "unbundled": [],
                    "bundledIds": [],
                },
            }
        ],
        "sentAt": f"{current_timestamp()}",
    }


def generate_payload(self: object, prompt: str) -> dict:
    """Creates conversation payload

    Args:
        self (object): _

    Returns:
        dict: _description_
    """

    payload_template = {
        "action": "next",
        "messages": [
            {
                # "id": "aaa2921d-d8c9-4516-80bd-ed1eaxxxxx",
                "author": {"role": "user"},
                "content": {"content_type": "text", "parts": [prompt]},
                "metadata": {},
            }
        ],
        # "conversation_id": self.conversation_metadata["id"],
        # "parent_message_id": "5b45a98c-0871-48ed-895b-f36f188cxxxx",
        "model": self.model,
        "timezone_offset_min": -180,
        "suggestions": [],
        "history_and_training_disabled": self.disable_history_and_training,
        # "arkose_token": None,
        "conversation_mode": {"kind": "primary_assistant"},
        "force_paragen": False,
        "force_rate_limit": False,
    }

    if self.current_conversation_id:
        # Continuing conversation
        payload_template["conversation_id"] = self.current_conversation_id
    else:
        # Create new conversation
        payload_template["messages"][0]["id"] = "aaa2" + uuid4().__str__()[4:]
        payload_template["suggestions"] = [
            prompt["prompt"] for prompt in self.prompt_library()["items"]
        ]

    # print(json.dumps( payload_template,indent=4,))
    return payload_template


def get_message(response: dict) -> str:
    """Extracts generated message from the response

    Args:
        response (dict): `bot.ask` response

    Returns:
        str: Extracted message
    """
    # print(json.dumps(response,indent=4))
    assert isinstance(response, dict), "'response' should be of 'dict' data-type"
    return response["message"]["content"]["parts"][0]
